/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * Version: 2.0b 09/20/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */ 
#ifndef _VTUN_H
#define _VTUN_H

#include "llist.h"

#define VER "2.0b"

/* Default VTUN port */
#define VTUN_PORT 5000

/* Default VTUN connect timeout in sec */
#define VTUN_CONNECT_TIMEOUT 30

/* General VTUN timeout for several operations, in sec */
#define VTUN_TIMEOUT 30

/* Default config file */
/* #define VTUN_CONFIG_FILE "/etc/vtund.conf" */

#define VTUN_ROUTE_COMMAND "/usr/sbin/route"
#define VTUN_FWALL_COMMAND "/usr/sbin/ipchains"

/* PID file */
/* #define VTUN_PID_FILE "/var/run/vtund.pid" */

/* End of configurable part */

struct vtun_stat {
   unsigned long byte_in;
   unsigned long byte_out;
   unsigned long comp_in;
   unsigned long comp_out;
};

struct vtun_cmd {
   char *prog;
   char *args;
   int  flags;
};
 
struct vtun_host {
   char *host;
   char *passwd;
   char *dev;

   llist up;
   llist down;

   int  flags;
   int  spd_in;
   int  spd_out;
   int  zlevel;

   int  rmt_fd;
   int  loc_fd;

   struct vtun_stat stat;
};

extern llist host_list;

/* Flags definitions */
#define VTUN_TTY       0x0100
#define VTUN_PIPE      0x0200
#define VTUN_ETHER     0x0400
#define VTUN_TYPE_MASK (VTUN_TTY | VTUN_PIPE | VTUN_ETHER) 

#define VTUN_TCP       0x0010  
#define VTUN_UDP       0x0020  
#define VTUN_PROT_MASK (VTUN_TCP | VTUN_UDP) 
#define VTUN_KEEPALIVE 0x0040	

#define VTUN_COMPRESS  0x0001
#define VTUN_SHAPE     0x0002
#define VTUN_ENCRYPT   0x0004

/* Mask to drop the flags which will be suplied by the server */
#define VTUN_CLNT_MASK 0xf000

/* Constants and flags for VTun protocol */
#define VTUN_FRAME_SIZE 2048

#define VTUN_FSIZE_MASK 0x0fff
#define VTUN_CONN_CLOSE 0x1000
#define VTUN_KEEP_ALIVE 0x2000

/* Authentification message size */
#define VTUN_MESG_SIZE 50

/* Global options */
struct vtun_opts {
   int server;
   int daemon;
   int port;
   int persist;
   int timeout;

   char *cfg_file;

   char *ppp;		 /* Comand to configure ppp devices */
   char *ifcfg;		 /* Comand to configure net devices */
   char *route;		 /* Comand to configure routing */
   char *fwall; 	 /* Comand to configure FireWall */

   unsigned long laddr;  /* Local ip address */
   char * sname;         /* Server's name address */
   unsigned long saddr;  /* Server's ip address */
};

extern struct vtun_opts vtun;

struct vtun_host * find_host(char *host);
int read_config(char *file);

void server(void);
void client(struct vtun_host *host);
#endif
